// extract.c last modified on: Mon Mar 17 16:38:47 CET 2008

// This is a replacement for extract.pl with useful enhancements:
// - can use different or multiple input/output delimiters
// - can extract or hide an undefined number of columns
// - can extract or hide a defined number of columns
// - can extract or hide a subset of the number of columns
// - can be easily modified
// - can be compiled and optimized for your host
// - about 4 times faster than perl version
// It is suggested to compile with:
// CFLAGS="-static -O3 -march=pentium4 -fomit-frame-pointer"
// gcc $CFLAGS -o extract extract.c
// gcc $CFLAGS -DSTRICT_MATCH -o extracts extract.c
// gcc $CFLAGS -DLOOSE_MATCH  -o extractl extract.c
// strip extract extracts extractl
 
#include <stdio.h>
#include <string.h>

// maximum row size
#define ROWSIZE 256
// token I/O delimiters
#define IDELIMITER ", "
#define ODELIMITER " "
// if defined, makes strict row pattern match:
// the number of colums from input must be
// equal to the number of executable arguments
// #define STRICT_MATCH 1
// if defined, match all but outputs only 'X' marked columns
// must be used in exclusive mode with STRICT_MATCH
// #define LOOSE_MATCH 1

// append string after string
// warning: you must check first for null pointers
inline void append(char* dst, const char* src){
 char temp[ROWSIZE];
 // check if dst is a null string
 if (dst[0]) {
  // dst is not null: use temp 
  snprintf(temp,ROWSIZE,"%s"ODELIMITER"%s",dst,src);
  // then copy to dst
  strcpy(dst,temp);
 // dst was a null string
 } else strcpy(dst,src);
}

int main(int argc, char** argv){
 int i;		// general purpose accumulator
 int virgin;	// =1 until first data out
 int space;	// '\n' counter 
 int xfound;	// number of 'X' found in argv
 int hfound;	// number of 'H' found in argv
 char outrow[ROWSIZE];	// used to compose output row

 // INITIAL CHECKS

 // we need at least one parameter
 if (argc<2) return 1;
 // check for 'X' and 'H' in argv
 xfound=0;
 hfound=0;
 for (i=1;i<argc;i++) {
  if (argv[i][0]=='X') xfound++;
  else if (argv[i][0]=='H') hfound++;
 }
 // cannot proceed without 'X' or 'H'
 if (!(xfound+hfound)) return 1;

 // PROCESS ROWS

 // output state initialization
 space=0;
 virgin=1;
 // process rows until EOF
 while (!feof(stdin)) {
  char row[ROWSIZE];	// store input row
  char* tok[ROWSIZE];	// store tokens
  int ntok;	// number of tokens in row
  int match;	// pattern match

  // get a row from stdin
  fgets(row,ROWSIZE,stdin);
  // if last char is not '\n' exit 
  // (this usually happens at the end of file)
  if (row[strlen(row)-1]!='\n') return 0;
  // check for null string
  if (row[0]=='\n'){
   // print only once
   if (space+1>=(argc-xfound-hfound)) {
    printf("\n");
    space-=(argc-xfound-hfound);
   }
   // we don't want space before data
   if (!virgin) space++;
   // skip other processing
   continue;
  }
  // cut last '\n' first to tokenize
  row[strlen(row)-1]=0;
  // if last row was data reset space
  space=0;
  // initialize tokens
  ntok=0;
  tok[0]=strtok(row,IDELIMITER);
  // collect tokens
  while (tok[ntok]) tok[++ntok]=strtok(0,IDELIMITER);

#ifdef STRICT_MATCH
  if (argc!=(ntok+1)) return 1;
#endif

  // initialize matching
  match=ntok;
  // initialize output row
  outrow[0]=0;
  // loop for every token
  for (i=0;i<ntok;i++){

#ifdef LOOSE_MATCH
   // skip if out of pattern or 'H' present
   if (((i+1)>=argc)||argv[i+1][0]=='H') continue;
   // for arguments marked with 'X' append token
   else if (argv[i+1][0]=='X') append(outrow,tok[i]);
#else
   // for arguments over argc or marked with 'X' append token
   if (((i+1)>=argc)||(argv[i+1][0]=='X')) append(outrow,tok[i]);
   // if marked with 'H' skip appending
   else if (argv[i+1][0]=='H') continue;
#endif

   // else check if strings are matching
   else if (strcmp(argv[i+1],tok[i])) match--;
  }
  // check if all fields matched
  if (ntok==match) {
   // output row
   printf("%s\n",outrow);
   // mark output started
   virgin=0;
  }
 }
 return 0;
}

